import java.io.File;

import be.ulb.db.Database;
import be.ulb.db.Record;
import be.ulb.db.RecordIterator;
import be.ulb.db.Relation;
import be.ulb.db.index.Index;
import be.ulb.db.schema.AttributeType;
import be.ulb.db.schema.AttributeTypeInt;
import be.ulb.db.schema.Schema;

public class IOTest {
    public static final int BLOCKSIZE  = 512;       // The size of a block
    public static final int NUMBLOCKS  = 1000;       // The number of blocks in the relation
    public static final int NUMBUFFERS = 100;       // The number of main memory buffers
    public static final int NUMRECORDS = 10000;      // The number of records to be stored

    private static final String DB_FILE      = "example.db"; //The file in which the database is stored

    private static final String FOO_RELATION = "Foo";      // The name of the Foo relation
    private static final String FOO_A        = "A";        // First attribute of the Foo relation
    private static final String FOO_B        = "B";        // Second attriute of the Foo relation
    private static final Schema FOO_SCHEMA;                // Schema of relation foo

    static {
            String [] names = { FOO_A, FOO_B };
            AttributeType [] types = { new AttributeTypeInt(),      // 4 bytes
                                       new AttributeTypeInt()       // 4 bytes
                                       };
            FOO_SCHEMA = new Schema(names, types);
    }

    private Database db                       = null; // The database

    /**
     * Utility function to print text to the screen
     */
    private static void log(String text) {
        System.out.println(text);
    }

    /**
     * Create a record, to be stored in the Foo relation, using the given
     * integer to fill in the fields. The A field is populated as i % 10,
     * while the B field is populated as i / 10.
     */
    private static Record createRecord(int i) {
        return null;
    }

    /**
     * Create a database, stored in the specified [[file]] with a relation named
     * Foo, and populate this relation.
     */
    private static void createDatabase(File file) throws Exception {
        log("Initializing the database");

        //create a database with NUMBLOCK blocks of BLOCKSIZE bytes, and a
        //buffermanager with NUMBUFFERS buffers
        Database db = new Database(file, NUMBLOCKS, BLOCKSIZE, NUMBUFFERS);
        Relation foo = new Relation(FOO_RELATION, FOO_SCHEMA, db.getCatalog(), db.getBufferManager());

        log("Loading " + NUMRECORDS + " records... ");
        //TODO: call createRecord repeatedly until NUMRECORDS have been loaded
        log("done");

        foo.close();
        db.close();

        System.gc();
        Thread.sleep(1);
    }

    /**
     * Get the file in which the database is expected to be written.
     */
    private static File getDatabaseFile() {
        return new File(DB_FILE);
    }

    /**
     * Returns a file with a valid database. If forceCreate is set to true,
     * a new database is generated. Otherwise, the existing database is reused.
     */
    private static File ensureDatabaseFile(boolean forceCreate) throws Exception {
        File file = getDatabaseFile();

        if (forceCreate && file.exists()) {
            file.delete();
        }

        if (!file.exists()) {
            createDatabase(file);
        }

        return file;
    }


    /**
     * Load the database from disk if it already exists, or create it otherwise.
     */
    private IOTest(boolean forceCreate) throws Exception {
        File dbfile = ensureDatabaseFile(forceCreate);
        db = new Database(dbfile, NUMBUFFERS);
    }

    /**
     * Retrieves the Relation named Foo from the database.
     */
    public Relation getFoo() throws Exception {
        return new Relation(FOO_RELATION, db.getCatalog(), db.getBufferManager());
    }

    /**
     * Print the number of I/Os that the database had to perform
     */
    private void printStatistics() throws Exception {
        // TODO: implement this.
        //   you may need to use db.getDiskManager()
        //   to retrieve the number of I/O
    }

    /**
     * Scan the given relation, printing every record.
     */
    public void scanRelation(Relation relation) throws Exception {
        final int SKIP_STEP = 500;  //Print one record out of every [SKIP_STEP]
        int count = 0;              //The number of records read so far

        log("Scanning foo:");
        RecordIterator iterator = relation.getIterator();

        while (iterator.hasNext()) {
            Record rec = iterator.next();

            //Only print a small portion of the results
            if (count % SKIP_STEP == 0) {
                if (count != 0) {
                    log("\t[skipped " + (SKIP_STEP - 1) + "]");
                }
                log("\t" + rec);
            }

            count++;
        }
        if (count % SKIP_STEP != 0) {
            log("\t[skipped " + (count % SKIP_STEP) + "]");
        }
        log("\t" + count + " records read");

        iterator.close();
    }

    /**
     * Initialize the database and run a few queries.
     */
    public static void main(String[] args) throws Exception{
        IOTest r = new IOTest(true);

        Relation foo = r.getFoo();

        log("Statistics before");
        r.printStatistics();

        r.scanRelation(foo);

        log("Statistics after");
        r.printStatistics();
    }
}
