package be.ulb.db;

import be.ulb.db.schema.*;
import java.nio.*;

/**
 * Manages a sequence of (record, info) pairs as slots in a block.
 */
public class SlotsManager {
    private int         slotSize;
    private ByteBuffer  buffer;
    private Schema      recordSchema;
    private Schema      infoSchema;
    private int         capacity;
    private int         end;

    public Schema getRecordSchema() {
        return recordSchema;
    }

    public Schema getInfoSchema() {
        return infoSchema;
    }

    /**
     * Allocate a temporary buffer that has the size of a slot
     */
    public ByteBuffer allocateTempSlot() {
        return ByteBuffer.allocate(slotSize);
    }

    /**
     * Allocate a temporary buffer that has the size of a record
     */
    public ByteBuffer allocateTempRecord() {
        return ByteBuffer.allocate(recordSchema.bytesize());
    }

    /**
     * Allocate a temporary buffer that has the size of slot information
     */
    public ByteBuffer allocateTempInfo() {
        return ByteBuffer.allocate(infoSchema.bytesize());
    }

    public SlotsManager(ByteBuffer buffer, Schema recordSchema, Schema infoSchema) {
        this.buffer       = buffer;
        this.recordSchema = recordSchema;
        this.infoSchema   = infoSchema;
        this.slotSize     = (recordSchema.bytesize() + infoSchema.bytesize());
        this.capacity     = buffer.capacity() / slotSize;
        this.end          = this.capacity * this.slotSize;
    }

    /**
     * Returns the schema of the keys.
     */
    public Schema getKeySchema() {
        return recordSchema;
    }

    /**
     * Returns the number of slots allocated.
     */
    public int capacity() {
        return this.capacity;
    }

    /**
     * Gets the location in the buffer corresponding to a given slot
     */
    private int slotLocation(int index) {
        return (index * slotSize);
    }

    /**
     * Reset the internal buffer in position and limit
     */
    private void resetBufferWindow() {
        ByteBufferUtils.resetBuffer(buffer);
    }

    /**
     * Position and limit the internal [buff] to the slot designated by [index]
     */
    private ByteBuffer windowBufferAt(int index, int beginOffset, int length) {
        int startPosition = slotLocation(index) + beginOffset;
        resetBufferWindow();
        buffer.position(startPosition);
        buffer.limit(startPosition + length);
        return buffer;
    }

    /**
     * Position and limit the internal [buff] to the slot designated by [index]
     */
    private ByteBuffer windowBufferAt(int index) {
        return windowBufferAt(index, 0, slotSize);
    }

    /**
     * Position and limit the internal [buff] to the record designated by [index]
     */
    private ByteBuffer windowBufferAtRecord(int index) {
        /*
        System.out.println("Windowing to record " + index);
        System.out.println(" -> Starting at " + slotLocation(index));
        System.out.println(" -> For a length of " + (slotSize - infoSchema.bytesize()));
        System.out.println(" -> Total buffer capacity " + buffer.capacity());
        */
        return windowBufferAt(index, 0, slotSize - infoSchema.bytesize());
    }

    /**
     * Position and limit the internal [buff] to the pointer designated by [index]
     */
    private ByteBuffer windowBufferAtInfo(int index) {
        return windowBufferAt(index, recordSchema.bytesize(), infoSchema.bytesize());
    }

    /**
     * Move a slot to another position from [[indexFrom]] to [[indexTo]],
     * using [[temp]] as a temporary storage for the moved slot.
     */
    public void moveSlot(int indexFrom, int indexTo, ByteBuffer temp) {
        windowBufferAt(indexFrom);

        ByteBufferUtils.resetBuffer(temp);
        temp.put(buffer);
        ByteBufferUtils.resetBuffer(temp);

        windowBufferAt(indexTo);
        buffer.put(temp);
        resetBufferWindow();
    }

    /**
     * Overwrite the first slots of this SlotsManager with the raw data
     * contained in [[data]]
     */
    public void bulkImport(ByteBuffer data) {
        if (data.capacity() % slotSize != 0) {
            //XXX: generate specific exception class
            throw new RuntimeException("Wrong data size for bulk import (request for " + data.capacity() + " while slots are " + slotSize + ")");
        }

        ByteBufferUtils.resetBuffer(data);
        ByteBufferUtils.resetBuffer(buffer);
        buffer.put(data);
    }

    /**
     * Write raw data over the first few slots.
     */
    public void setHead(ByteBuffer data) {
        //XXX: throw a more specific exception
        if (data.capacity() % slotSize != 0) throw new RuntimeException("Wrong slot size");
        ByteBufferUtils.resetBuffer(buffer);
        buffer.put(data);
    }

    /**
     * Get a raw representation of the slots in the range [overflowFrom, end)
     */
    public ByteBuffer getTail(int from) {
        int slotstart = slotLocation(from);
        return ByteBufferUtils.resetBuffer(buffer, slotstart, end).slice();
    }

    /**
     * Get the byte buffer where the information attached to a given slot can be
     * found.
     */
    public ByteBuffer getInfoBuffer(int index) {
        return windowBufferAtInfo(index).slice();
    }

    /**
     * Read the information attached to the given slot.
     */
    public Record getInfo(int index) {
        return new Record(infoSchema, getInfoBuffer(index));
    }

    /**
     * Set the information attached to the given slot.
     */
    public void putInfo(int index, Record info) {
        windowBufferAtInfo(index).put(info.getByteBuffer());
    }

    /**
     * Read the record in the given slot. If a buffer is specified, the
     * raw data of the records are first copied to the buffer, and this buffer
     * is used. Otherwise, the buffer is shared with that managed by the
     * slots manager.
     */
    private Record readRecord(int index, ByteBuffer buffer) {
        ByteBuffer recordBuffer = windowBufferAtRecord(index).slice();
        resetBufferWindow();

        if (buffer != null) {
            //Copy the record to the target buffer
            buffer.put(recordBuffer);
            recordBuffer = buffer;
        }

        // Read the record
        return new Record(recordSchema, recordBuffer);
    }

    /**
     * Read the record in the given slot. The memory allocated
     * for this record, is shared with the SlotsManager buffer.
     */
    public Record readRecordShared(int index) {
        return readRecord(index, null);
    }

    /**
     * Read the record in the given slot. A new buffer is allocated
     * for this record, avoiding modification to the original data.
     */
    public Record readRecord(int index) {
        return readRecord(index, allocateTempRecord());
    }

    /**
     * Set the record at the given slot to the given values.
     */
    public void putRecord(int index, Record rec) {
        windowBufferAtRecord(index).put(rec.getByteBuffer());
    }

    /**
     * Set the record and information at the given slot to the given values.
     */
    public void putPair(int index, Record rec, Record info) {
        putRecord(index, rec);
        putInfo(index, info);
    }
}

