/*-------------------------------------------------------------------
   The BTIndex class provides the interface to an implementation of the BTree index
   data structure.

   $Id: BTIndex.java,v 1.1.1.1 2005/09/26 08:19:41 stijn Exp $
 -------------------------------------------------------------------*/
package be.ulb.db.index.btree;

import java.util.*;
import java.io.*;

import be.ulb.db.*;
import be.ulb.db.buffermanager.*;
import be.ulb.db.diskmanager.*;
import be.ulb.db.index.*;
import be.ulb.db.schema.*;

/**
 * An accessor to a B+Tree index.
 */
public class BTIndex extends Index
{
    BlockAddress fixedRoot;
    List<BTIndexIterator> iterators;

    private void setupFixedRoot(Schema keySchema, BufferManager bm) throws BufferManagerException, DiskManagerException, IOException {
        BTIndexContainerFactory factory = new BTIndexContainerFactory(keySchema, bm);
        BTIndexContainer container = bm.newBlock(factory, BTIndexContainer.class);
        fixedRoot = container.getBlockAddress();
        bm.unpinBlock(container);
    }

    private BTIndexContainer loadFixedRoot(Schema keySchema, BufferManager bm) throws BufferManagerException, IOException, DiskManagerException {
        BTIndexContainerFactory factory = new BTIndexContainerFactory(keySchema, bm);
        return (BTIndexContainer)bm.pinBlock(fixedRoot, factory);
    }

    private void invalidateIterators() {
        for (BTIndexIterator iterator: iterators) {
            iterator.invalidate();
        }

        iterators.clear();
    }

    /* Creates a new BT index */
    public BTIndex(Schema keySchema, BufferManager bm)
    {
        super(keySchema, bm);
        iterators = new ArrayList<BTIndexIterator>();

        try {
            setupFixedRoot(keySchema, bm);
        }
        catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    /** Load the index which starts at address addr.  */
    public BTIndex(Schema keySchema, BufferManager bm, BlockAddress addr)
    {
        super(keySchema, bm);
        iterators = new ArrayList<BTIndexIterator>();
        fixedRoot = addr;
    }

    /** destroy the index - free all blocks belonging to the index */
    public void destroy()
    {
        close();

        try {
            BTIndexContainer root = loadFixedRoot(keySchema, bm);
            root.destroy();
            bm.freeBlock(root);
            fixedRoot = BlockAddress.INVALIDADDR;
        }
        catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    /* Close the index and release all resources. (i.e. Unpin all
     * buffers and close all open iterators if any) */
    public void close()
    {
        invalidateIterators();
    }

    /** returns the blockaddress from which the index can be loaded
     * next time. This address has to be fixed over time! (the root of
     * the BTree will not do, as this can change!) */
    public BlockAddress getBlockAddress()
    {
        return fixedRoot;
    }

    /** Add RecordAddress addr to the BTree, knowing that the record
     * at addr equals keyvalue on its attributes in keySchema.  */
    public void addRecord(Record keyvalue, RecordAddress addr)
    {
        invalidateIterators();
        super.addRecord(keyvalue, addr);
        try {
            BTIndexContainer root = loadFixedRoot(keySchema, bm);
            root.addRecord(keyvalue, addr);
            bm.unpinBlock(root);
        }
        catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    /** Search in the index for RecordAddresses whose corresponding
     *  records equal keyvalue in their keySchema attributes.  The
     *  recordIterator should return records of type keySchema */
    public RecordAddressIterator findRecord(Record key)
    {
        BTIndexIterator result = null;
        //System.out.println("Looking for key " + key + " in index");

        super.findRecord(key);
        try {
            BTIndexContainer root = loadFixedRoot(keySchema, bm);
            result = root.findRecord(key);
            iterators.add(result);
            bm.unpinBlock(root);
        }
        catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
        return result;
    }

    /** Delete the record at addr, whose Key value is keyvalue. */
    public void removeRecord(Record keyvalue, RecordAddress addr)
    {
        super.removeRecord(keyvalue, addr);
        throw new RuntimeException("Not Implemented");
    }

    public String toString() {
        String result = "";
        try {
            BTIndexContainer root = loadFixedRoot(keySchema, bm);
            result = root.toString();
            bm.unpinBlock(root);
        }
        catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }

        return result;
    }
}
