/**
 * The superclass for blocks of a B+Tree.
 */
package be.ulb.db.index.btree;

import java.io.IOException;
import java.nio.ByteBuffer;

import be.ulb.db.Block;
import be.ulb.db.BlockAddress;
import be.ulb.db.BlockFullException;
import be.ulb.db.ByteBufferUtils;
import be.ulb.db.DuplicateKeyException;
import be.ulb.db.Record;
import be.ulb.db.RecordAddress;
import be.ulb.db.SlotsManager;
import be.ulb.db.buffermanager.BufferManager;
import be.ulb.db.buffermanager.BufferManagerException;
import be.ulb.db.diskmanager.DiskManagerException;
import be.ulb.db.schema.AttributeType;
import be.ulb.db.schema.AttributeTypeBlockAddress;
import be.ulb.db.schema.AttributeTypeBoolean;
import be.ulb.db.schema.AttributeTypeInt;
import be.ulb.db.schema.Schema;

/**
 * A B-Tree index block in which all records are sorted.
 *
 * @param AddressType The type of an address pointer (RecordAddress for leaves, BlockAddress for inner)
 * @param AddressAttributeType The attribute type corresponding to the * AddressType
 */
public abstract class BTIndexBlock extends Block {
    /**
     * Specifies a record corresponding to the header of a [[BTIndexBlock]]
     */
    protected static class HeaderRecord extends Record {
        public static int SIZE_OFFSET = 0;
        public static int LEAF_OFFSET = 1;
        public static int EPTR_OFFSET = 2;
        public static Schema SCHEMA = Schema.createAnonymousSchema(new AttributeType[] {
            new AttributeTypeInt(),               // Size
            new AttributeTypeBoolean(),           // Is leaf?
            new AttributeTypeBlockAddress()       // Additional pointer
        });

        public static int SIZE = SCHEMA.bytesize();

        /**
         * Construct a header record on the specified buffer.
         */
        public HeaderRecord(ByteBuffer buffer) {
            super(SCHEMA, buffer);
        }

        /**
         * Get the number of entries stored in the index block.
         */
        public int getSize()                 { return getInt(SIZE_OFFSET); }

        /**
         * Set the number of entries stored in the index block.
         */
        public void setSize(int newValue)    { put(SIZE_OFFSET, newValue); }

        /**
         * Get whether the block is a leaf (True) or an inner block (False).
         */
        public boolean getType()             { return getBoolean(LEAF_OFFSET); }

        /**
         * Set whether the block is a leaf (True) or an inner block (False).
         */
        public void setType(boolean newValue) { put(LEAF_OFFSET, newValue); }

        /**
         * Get the additional pointer.
         */
        public BlockAddress getPtr()         { return getBlockAddress(EPTR_OFFSET); }

        /**
         * Set the additional pointer.
         */
        public void setPtr(BlockAddress newValue) { put(EPTR_OFFSET, newValue); }
    }

    /**
     * The result of looking for a key in the current block
     */
    protected static class FindKeyInfo {
        private Record key;
        private boolean foundKey;
        private int upperIndex;

        public boolean found() { return foundKey; }
        public int getLowerIndex() { return foundKey? upperIndex : upperIndex - 1; }
        public int getUpperIndex() { return upperIndex; }
        public Record getKey() { return key; }

        public FindKeyInfo(Record key, int upperIndex, boolean found) {
            this.key = key;
            this.upperIndex = upperIndex;
            this.foundKey = found;
        }
    }

    /**
     * Describes how a split occurred.
     */
    public class SplitInformation {
        private BTIndexBlock newChild;
        private Record splitKey;

        /**
         * Indicates whether a split occurred or not.
         */
        public boolean splitted() { return newChild != null; }

        /**
         * Retrieves the block that was created as the result of splitting.
         */
        public BTIndexBlock getNewBlock() { return newChild; }

        /**
         * Retrieves the address of the new block that was created as the result of splitting.
         */
        public BlockAddress getNewBlockAddress() { return newChild.getBlockAddress(); }

        /**
         * Retrieves the first key that correspond to the new block that was
         * created.
         */
        public Record getSplitKey() { return splitKey; }

        /**
         * Construct a new SplitInformation describing how a split occurred.
         * If no split occurred [[newChild]] and [[splitKey]] must both be null.
         */
        public SplitInformation(BTIndexBlock newChild, Record splitKey) {
            this.newChild = newChild;
            this.splitKey = splitKey;
        }
    }

    protected boolean isDirty;          // Whether we have changes not yet on the disk
    protected SlotsManager manager;     // Management for slots inside this block
    protected HeaderRecord header;      // An accessor to the header record of this block

    private BTIndexBlockFactory blockFactory; // A factory that revives blocks from the disk

    /**
     * Gets a factory that reads blocks from the disk. This function is used with
     * pinblock to load a B+Tree block from the disk.
     */
    protected BTIndexBlockFactory getBlockFactory() {
        if (blockFactory == null)
            blockFactory = new BTIndexBlockFactory(manager.getRecordSchema(), BTIndexBlockFactory.BlockInstanciationMode.FromRecord);
        return blockFactory;
    }

    /**
     * Gets the schema of the information associated with each entry of this
     * block. For instance, leaf blocks need to store a RecordAddress, pointing
     * to the record in the main relation, along with each entry.
     */
    protected abstract Schema getInfoSchema();

    /**
     * Destroy the B+Tree
     */
    public abstract void destroy(BufferManager bm);

    /**
     * Indicates whether the block is a leaf
     */
    public abstract boolean isLeaf();

    /**
     * Split this block in such a way that the newly created block contains
     * about half the slots of this block, and that the number of slots in this
     * block is halved.
     */
    public abstract SplitInformation split(BufferManager bm) throws IOException, BufferManagerException, DiskManagerException;

    /**
     * Determines whether the block whose content is in [[buffer]] is a leaf.
     */
    public static boolean isLeafBlock(ByteBuffer buffer) {
        HeaderRecord r = bindHeader(buffer);
        return r.getType();
    }

    /**
     * Get a header record mapped to the specified buffer.
     */
    private static HeaderRecord bindHeader(ByteBuffer buffer) {
        return new HeaderRecord(ByteBufferUtils.resetBuffer(buffer, HeaderRecord.SIZE).slice());
    }

    /**
     * Create a new BTIndexBlock, storing sorted records of the specified
     * [keySchema] in a block stored at the specified [address] and whose content
     * can be accessed through [buffer].
     */
    public BTIndexBlock(Schema keySchema, ByteBuffer buffer, BlockAddress address) {
        super(buffer, address);

        this.isDirty = false;

        this.manager = new SlotsManager(
                ByteBufferUtils.resetBuffer(
                    buffer, HeaderRecord.SIZE, buffer.capacity() - HeaderRecord.SIZE).slice(),
                keySchema, getInfoSchema());

        //Link up our header record to the first few bytes of the block
        this.header = bindHeader(buffer);
    }

    /**
     * Indicates whether the block has changed since it was read from disk.
     */
    public boolean isModified() {
        return isDirty;
    }

    /**
     * Indicates that this block has been modified.
     */
    private void setModified() {
        isDirty = true;
    }

    /**
     * Initialize the content of this block
     */
    public void initContents() {
        // Write our header. Initially, the block is empty.
        header.setSize(0);
        header.setType(isLeaf());
        header.setPtr(BlockAddress.INVALIDADDR);
        setModified();

        //System.out.println("*** block created ***");
    }

    /**
     * Finds how many records we can store in this block
     */
    public int capacity() {
        return manager.capacity();
    }

    /**
     * Gets how many records are currently stored in this block
     */
    public int size() {
        return header.getSize();
    }

    /**
     * Determines whether this block can hold new records
     */
    public boolean isFull() {
        return size() >= manager.capacity();
    }

    /**
     * Sets how many records are currently stored in this block
     */
    protected void setSize(int newSize) {
        header.setSize(newSize);
        setModified();
    }

    /**
     * Set the initial data contained in this node from raw data of another
     * block.
     */
    protected void receiveOverflow(int size, ByteBuffer overflow) {
        manager.setHead(overflow);
        setSize(size);
    }

    /**
     * Move part of the data at the end of this block to another block.
     */
    protected void transferOverflow(int size, BTIndexBlock target) {
        int oldSize = size();
        int newSize = oldSize - size;

        target.receiveOverflow(size, manager.getTail(oldSize - size));
        setSize(newSize);
    }

    /**
     * Add a key into this block, splitting this block if necessary.
     */
    protected SplitInformation localAddRecord(Record key, Record slotinfo, BufferManager buffermanager) throws DuplicateKeyException, IOException, BufferManagerException, DiskManagerException {
        SplitInformation info = new SplitInformation(null, null);
        BTIndexBlock insertionTarget = this;

        // Cannot simply add an item; split ourself.
        if (isFull()) {
            info = split(buffermanager);

            if (key.compareTo(info.getSplitKey()) >= 0) {
                insertionTarget = info.getNewBlock();
            }
        }

        // Actually insert the new record
        try { insertionTarget.localAddRecord(key, slotinfo); }
        catch (BlockFullException e) {
            //Should not occur, since we just split.
            e.printStackTrace();
            System.exit(1);
        }

        return info;
    }

    /**
     * Add a record into the B+Tree, recursively
     */
    public abstract SplitInformation addRecord(Record key, RecordAddress address, BufferManager buffermanager)
        throws DuplicateKeyException, IOException, DiskManagerException, BufferManagerException;

    /**
     * Find the given key in the slots of this block.
     */
    protected FindKeyInfo findKey(Record key)
    {
        int length = size();
        int start = 0;

        while (length != 0)
        {
            //System.out.println("Checking the range [" + start + ", " + (start + length) + ")");
            int newlength = length / 2;
            int checkindex = start + newlength; // Check at mid range

            Record existing = manager.readRecordShared(checkindex);
            int compared = (existing.compareTo(key));
            //System.out.println("Comparison result at start: " + compared);

            if (compared == 0) {
                return new FindKeyInfo(key, checkindex, true);
            }
            else if (compared < 0) {
                start += newlength + length % 2;
            }

            length = newlength;
            //System.out.println("length <- " + length);
            //System.out.println("start <- " + start);
        }

        return new FindKeyInfo(key, start, false);
    }

    /**
     * Insert the given record and information at the specified location,
     * moving any existing record in the range [index, max) to [index + 1, max].
     */
    private void insertPair(int index, Record key, Record info) {
        //System.out.println(String.format("Inserting record %s at %d", rec, index));
        ByteBuffer temp = manager.allocateTempSlot();

        int max = size();
        while (max != index) {
            manager.moveSlot(max - 1, max, temp);
            max--;
        }

        manager.putPair(index, key, info);
        header.setSize(size() + 1);
        setModified();
    }


    /**
     * Ensures that the
     */
    private void checkIndex(int index) throws IndexOutOfBoundsException {
        if (index < 0 || index >= size())
            throw new IndexOutOfBoundsException("Index " + index + " does not designate a currently allocated slot");
    }

    /**
     * Add a new key,address pair into this block at the right slot.
     *
     * @throws BlockFullException when the block is full.
     * @throws DuplicateKeyException if the key is already stored in the current block.
     */
    protected void localAddRecord(Record key, Record info) throws BlockFullException, DuplicateKeyException {
        if (isFull()) throw new BlockFullException("Block is full (" + size() + " record / " + capacity() + " slots)");

        // Find the location where the new record will be inserted
        FindKeyInfo index = findKey(key);

        if (index.found()) throw new DuplicateKeyException();
        assert(index.getUpperIndex() < size());

        //Insert the record right at the index
        insertPair(index.getUpperIndex(), key, info);
    }

    /**
     * Remove the entry at the specified index
     */
    public void removeRecord(int index) {
        checkIndex(index);

        int last = size() - 1;
        ByteBuffer buffer = manager.allocateTempSlot();

        while (index != last) {
            manager.moveSlot(index + 1, index, buffer);
            index++;
        }

        header.setSize(size() - 1);

        setModified();
    }

    public Record getRecord(int index) {
        checkIndex(index);
        return manager.readRecord(index);
    }

    public Record getFirstRecord() {
        return getRecord(0);
    }

    public abstract BTIndexIterator findRecord(BufferManager bm, Record key);

    public String toString() {
        String result = "[" + (isLeaf() ? "Leaf" : "Inner") + " BTIndexBlock of " + size() + " records:\n";

        if (!isLeaf()) result += "\t -1: -> " + header.getPtr() + "\n";
        for (int i = 0; i < size(); ++i) {
            result += "\t " + i + ":" + manager.readRecordShared(i).toString();
            result += " -> " + manager.getInfo(i).toString();
            result += "\n";
        }

        result += "]";

        return result;
    }

    public abstract String toString(BufferManager bm, int indent);

    public String toString(BufferManager bm) {
        return toString(bm, 0);
    }
}
