package be.ulb.db.index.btree;

import java.io.IOException;
import java.nio.ByteBuffer;

import be.ulb.db.BlockAddress;
import be.ulb.db.DuplicateKeyException;
import be.ulb.db.Record;
import be.ulb.db.RecordAddress;
import be.ulb.db.buffermanager.BufferManager;
import be.ulb.db.buffermanager.BufferManagerException;
import be.ulb.db.diskmanager.DiskManagerException;
import be.ulb.db.schema.AttributeType;
import be.ulb.db.schema.AttributeTypeBlockAddress;
import be.ulb.db.schema.Schema;

class BTIndexBlockInner extends BTIndexBlock {
    /**
     * Specifies the address of a record corresponding to a key
     */
    private static class RecordPointer extends Record {
        public static int PTR_OFFSET = 0;
        public static Schema SCHEMA = Schema.createAnonymousSchema(new AttributeType[] { new AttributeTypeBlockAddress() });

        public BlockAddress getPtr()       { return getBlockAddress(PTR_OFFSET); }
        public void setPtr(BlockAddress newValue) { put(PTR_OFFSET, newValue); }

        /**
         * Construct a pointer record on the specified buffer.
         */
        public RecordPointer(ByteBuffer buffer) {
            super(SCHEMA, buffer);
        }

        /**
         * Construct new a pointer record from an address
         */
        public RecordPointer(BlockAddress address) {
            super(SCHEMA);
            setPtr(address);
        }
    }

    private BTIndexBlockFactory innerFactory;

    /**
     * Lazy-instanciation for the inner blocks factory.
     */
    protected BTIndexBlockFactory getNewBlockFactory() {
        if (innerFactory == null)
            innerFactory = new BTIndexBlockFactory(manager.getRecordSchema(), BTIndexBlockFactory.BlockInstanciationMode.Inner);
        return innerFactory;
    }

    public boolean isLeaf() { return false; }
    protected Schema getInfoSchema() { return RecordPointer.SCHEMA; }

    private BlockAddress getChildAddress(int index) {
        BlockAddress address = header.getPtr();

        if (index != -1) {
            ByteBuffer buff = manager.getInfoBuffer(index);
            RecordPointer rp = new RecordPointer(buff);
            address = rp.getPtr();
        }

        return address;
    }

    private BTIndexBlock pinChild(BufferManager buffermanager, int index) throws IOException, DiskManagerException, BufferManagerException {
        BlockAddress address = getChildAddress(index);
        return (BTIndexBlock)buffermanager.pinBlock(address, getBlockFactory());
    }

    /**
     * Destroy the B+Tree
     */
    public void destroy(BufferManager bm) {
        for (int i = -1; i != size(); ++i) {
            try {
            BTIndexBlock child = pinChild(bm, i);
            child.destroy(bm);
            bm.freeBlock(child);
            }
            catch (Exception e) {
                e.printStackTrace();
                System.exit(1);
            }
        }
    }

    private int getChildForKey(Record key) {
        FindKeyInfo info = findKey(key);
        return info.getLowerIndex();
    }

    private BTIndexBlock pinChildForKey(BufferManager buffermanager, Record key)
        throws IOException, DiskManagerException, BufferManagerException {
        return pinChild(buffermanager, getChildForKey(key));
    }

    /**
     * Initialize this node as the new root of a B+Tree.
     */
    public void initializeAsRoot(Record key, BTIndexBlock left, BTIndexBlock right) {
        assert(size() == 0);
        assert(left.getBlockAddress() != BlockAddress.INVALIDADDR);
        assert(right.getBlockAddress() != BlockAddress.INVALIDADDR);
        header.setPtr(left.getBlockAddress());
        try { localAddRecord(key, new RecordPointer(right.getBlockAddress())); }
        catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    public SplitInformation addRecord(Record key, RecordAddress address, BufferManager buffermanager) throws DuplicateKeyException, IOException, BufferManagerException, DiskManagerException {
        SplitInformation info = new SplitInformation(null, null);
        BTIndexBlock childblock = pinChildForKey(buffermanager, key);

        // Add the pair down the tree. The [[childInfo]] variable
        // indicates whether a new child node has been created that need to
        // be registered at this level.
        try {
            SplitInformation childInfo = childblock.addRecord(key, address, buffermanager);

            if (childInfo.splitted()) {
                Record pointer = new RecordPointer(childInfo.getNewBlockAddress());
                info = localAddRecord(childInfo.getSplitKey(), pointer, buffermanager);
                buffermanager.unpinBlock(childInfo.getNewBlock());
            }
        }
        finally {
            buffermanager.unpinBlock(childblock);
        }

        return info;
    }

    protected void setFirstPointer(BlockAddress ptr) {
        header.setPtr(ptr);
    }

    public SplitInformation split(BufferManager bm) throws IOException, BufferManagerException, DiskManagerException {
        BTIndexBlockInner sibling = bm.newBlock(getNewBlockFactory(), BTIndexBlockInner.class);

        //Transfer half the records to the new block
        int siblingSize = size() / 2;
        transferOverflow(siblingSize, sibling);

        //Move one more record as follows:
        // - move the key up in the tree
        // - move the pointer as the first pointer of the sibling
        int last = size() - 1;
        Record splitKey = manager.readRecord(last);
        BlockAddress ptr = getChildAddress(last);
        sibling.setFirstPointer(ptr);
        setSize(last); //Remove the block from here

        return new SplitInformation(sibling, splitKey);
    }

    public BTIndexIterator findRecord(BufferManager bm, Record key) {
        BTIndexIterator result = null;
        try {
            FindKeyInfo info = findKey(key);
            BTIndexBlock block = pinChild(bm, info.getLowerIndex());
            result = block.findRecord(bm, key);
            bm.unpinBlock(block);
        }
        catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
        return result;
    }

    public BTIndexBlockInner(Schema recordSchema, ByteBuffer buffer, BlockAddress address) {
        super(recordSchema, buffer, address);
        innerFactory = null;
    }

    private String indent(int indent) {
        String result = "";
        while (indent-- != 0) result += "\t";
        return result;
    }

    public String toString(BufferManager bm, int indent) {
        String ind = this.indent(indent);
        String result = ind + "[Inner of " + size() + " records, @" + getBlockAddress() + ":\n";

        result += ind + "\t -1: (first block) -> \n";
        try {
            BTIndexBlock block = pinChild(bm, -1);
            result += block.toString(bm, indent + 2) + "\n";
            bm.unpinBlock(block);
        }
        catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }

        for (int i = 0; i < size(); ++i) {
            result += ind + "\t " + i + ":" + manager.readRecordShared(i).toString() + " -> \n";
            try {
                BTIndexBlock block = pinChild(bm, i);
                result += block.toString(bm, indent + 2) + "\n";
                bm.unpinBlock(block);
            }
            catch(Exception e) {
                e.printStackTrace();
                System.exit(1);
            }
        }

        result += ind + "]";

        return result;
    }
}
