package be.ulb.db.index.btree;

import java.io.IOException;
import java.nio.ByteBuffer;

import be.ulb.db.BlockAddress;
import be.ulb.db.DuplicateKeyException;
import be.ulb.db.Record;
import be.ulb.db.RecordAddress;
import be.ulb.db.buffermanager.BufferManager;
import be.ulb.db.buffermanager.BufferManagerException;
import be.ulb.db.diskmanager.DiskManagerException;
import be.ulb.db.schema.AttributeType;
import be.ulb.db.schema.AttributeTypeRecordAddress;
import be.ulb.db.schema.Schema;

public class BTIndexBlockLeaf extends BTIndexBlock {
    /**
     * Specifies the address of a record corresponding to a key
     */
    private static class RecordPointer extends Record {
        public static int PTR_OFFSET = 0;
        public static Schema SCHEMA = Schema.createAnonymousSchema(new AttributeType[] { new AttributeTypeRecordAddress() });

        public RecordAddress getPtr()       { return getRecordAddress(PTR_OFFSET); }
        public void setPtr(RecordAddress newValue) { put(PTR_OFFSET, newValue); }

        /**
         * Construct a pointer record on the specified buffer.
         */
        public RecordPointer(ByteBuffer buffer) {
            super(SCHEMA, buffer);
        }

        /**
         * Construct a pointer record pointing to the given address
         */
        public RecordPointer(RecordAddress address) {
            super(SCHEMA);
            setPtr(address);
        }
    }

    private BTIndexBlockFactory leafFactory;

    protected Schema getInfoSchema() {
        return RecordPointer.SCHEMA;
    }

    private void setNextPtr(BlockAddress address) {
        header.setPtr(address);
    }

    public BlockAddress getNextPtr() {
        return header.getPtr();
    }

    public boolean hasNextSibling() {
        return !getNextPtr().equals(BlockAddress.INVALIDADDR);
    }

    public BTIndexBlockLeaf getNextSibling(BufferManager bm) throws IOException, BufferManagerException, DiskManagerException {
        return (BTIndexBlockLeaf)bm.pinBlock(getNextPtr(), getBlockFactory());
    }

    public RecordAddress getRecordAddressAt(int slot) {
        RecordPointer rec = new RecordPointer(manager.getInfoBuffer(slot));
        return rec.getPtr();
    }

    /**
     * Destroy the B+Tree
     */
    public void destroy(BufferManager bm) { /* Nothing to do */ }

    /**
     * Lazy-instantiation for the leaf blocks factory.
     */
    protected BTIndexBlockFactory getNewBlockFactory() {
        if (leafFactory == null)
            leafFactory = new BTIndexBlockFactory(manager.getRecordSchema(), BTIndexBlockFactory.BlockInstanciationMode.Leaf);
        return leafFactory;
    }

    public SplitInformation split(BufferManager bm) throws IOException, BufferManagerException, DiskManagerException {
        BTIndexBlockLeaf sibling = bm.newBlock(getNewBlockFactory(), BTIndexBlockLeaf.class);

        //Transfer half the records to the new block
        int siblingSize = size() / 2;
        transferOverflow(siblingSize, sibling);

        //Chain the leaf blocks
        sibling.setNextPtr(getNextPtr());
        setNextPtr(sibling.getBlockAddress());

        return new SplitInformation(sibling, sibling.getFirstRecord());
    }

    public boolean isLeaf() { return true; }

    public SplitInformation addRecord(Record key, RecordAddress address, BufferManager buffermanager) throws DuplicateKeyException, IOException, BufferManagerException, DiskManagerException {
        return localAddRecord(key, new RecordPointer(address), buffermanager);
    }

    public int findRecordPosition(Record key) {
        FindKeyInfo info = findKey(key);
        return info.found() ? info.getUpperIndex() : -1;
    }

    private BTIndexBlockLeaf duplicate(BufferManager bm) throws IOException, BufferManagerException, DiskManagerException {
        BTIndexBlockLeaf result = (BTIndexBlockLeaf)bm.pinBlock(this.getBlockAddress(), getNewBlockFactory());
        return result;
    }

    public BTIndexIterator findRecord(BufferManager bm, Record key) {
        FindKeyInfo info = findKey(key);
        try {
            return new BTIndexIterator(bm, duplicate(bm), info.getUpperIndex());
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);

            return null;
        }
    }

    /**
     * Create a new BTIndexBlock, storing the leaf of a B+Tree
     */
    public BTIndexBlockLeaf(Schema keySchema, ByteBuffer buffer, BlockAddress address) {
        super(keySchema, buffer, address);
        leafFactory = null;
    }

    private String indent(int width) {
        String result = "";
        while (width-- != 0) result += "\t";
        return result;
    }

    public String toString(BufferManager bm, int indent) {
        String ind = this.indent(indent);
        String result = ind + "[Leaf of " + size() + " records:\n";

        for (int i = 0; i < size(); ++i) {
            result += ind + "\t" + manager.readRecordShared(i).toString() + " -> " + manager.getInfo(i).toString() + "\n";
        }

        result += ind + "]";

        return result;
    }
}
