package be.ulb.db.index.btree;

import java.io.IOException;
import java.nio.ByteBuffer;

import be.ulb.db.Block;
import be.ulb.db.BlockAddress;
import be.ulb.db.ByteBufferUtils;
import be.ulb.db.DuplicateKeyException;
import be.ulb.db.Record;
import be.ulb.db.RecordAddress;
import be.ulb.db.buffermanager.BufferManager;
import be.ulb.db.buffermanager.BufferManagerException;
import be.ulb.db.diskmanager.DiskManagerException;
import be.ulb.db.schema.AttributeType;
import be.ulb.db.schema.AttributeTypeBlockAddress;
import be.ulb.db.schema.AttributeTypeInt;
import be.ulb.db.schema.Schema;
import be.ulb.db.schema.SchemaException;

/**
 * A block implementing a fixed root for a B+Tree
 */
public class BTIndexContainer extends Block {
    protected static class HeaderRecord extends Record {
        public static int RPTR_OFFSET = 0;
        public static int SIZE_OFFSET = 1;
        public static Schema SCHEMA = Schema.createAnonymousSchema(new AttributeType[] {
            new AttributeTypeBlockAddress(),        // Pointer to root
            new AttributeTypeInt()                  // Total number of keys
        });

        public static int SIZE = SCHEMA.bytesize();

        /**
         * Construct a header record on the specified buffer.
         */
        public HeaderRecord(ByteBuffer buffer) {
            super(SCHEMA, buffer);
        }

        /**
         * Get the block in which the root of the B+Tree is located.
         */
        public BlockAddress getRoot()              { return getBlockAddress(RPTR_OFFSET); }

        /**
         * Set the block in which the root of the B+Tree is located.
         */
        public void setRoot(BlockAddress newValue) { put(RPTR_OFFSET, newValue); }

        /**
         * Get the number of (key, recordAddress) pairs currently stored in the
         * B+Tree
         */
        public int getSize()                       { return getInt(SIZE_OFFSET); }

        /**
         * Set the number of (key, recordAddress) pairs currently stored in the
         * B+Tree
         */
        public void setSize(int newValue)          { put(SIZE_OFFSET, newValue); }
    }

    private boolean isDirty;
    private Schema keySchema;
    private BufferManager bm;
    private HeaderRecord header;
    private BTIndexBlockFactory loadFactory;
    private BTIndexBlockFactory innerFactory;

    /**
     * Indicates whether the block has changed since it was read from disk.
     */
    public boolean isModified() {
        return isDirty;
    }

    /**
     * Indicates that this block has been modified.
     */
    private void setModified() {
        isDirty = true;
    }

    /**
     * Construct a HeaderRecord bound to the specified [[buffer]]
     */
    private HeaderRecord bindHeader(ByteBuffer buffer) {
        return new HeaderRecord(ByteBufferUtils.resetBuffer(buffer, HeaderRecord.SIZE).slice());
    }

    public BTIndexContainer(Schema keySchema, ByteBuffer buffer, BlockAddress address, BufferManager bm) {
        super(buffer, address);
        this.isDirty = false;
        this.keySchema = keySchema;
        this.bm = bm;
        this.loadFactory = new BTIndexBlockFactory(keySchema, BTIndexBlockFactory.BlockInstanciationMode.FromRecord);
        this.innerFactory = new BTIndexBlockFactory(keySchema, BTIndexBlockFactory.BlockInstanciationMode.Inner);
        this.header = bindHeader(buffer);
    }

    /**
     * Create a leaf node as the first root of the B+Tree.
     */
    private BTIndexBlock createInitialRoot() throws BufferManagerException, DiskManagerException, IOException {
        BTIndexBlockFactory factory = new BTIndexBlockFactory(keySchema, BTIndexBlockFactory.BlockInstanciationMode.Leaf);
        return (BTIndexBlock)bm.newBlock(factory);
    }

    /**
     * Set the address of the real B+Tree root
     */
    private void setRoot(Block block) {
        header.setRoot(block.getBlockAddress());

        setModified();
    }

    /**
     * Initialize the B+Tree with a single leaf node.
     */
    public void initContents() {
        // Write our header. Initially, the root is a leaf
        try {
            BTIndexBlock root = createInitialRoot();
            setRoot(root);
            setSize(0);
            bm.unpinBlock(root);
        }
        catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    /**
     * Pin the root from the given address
     */
    private BTIndexBlock pinRoot() throws IOException, BufferManagerException, DiskManagerException {
        return (BTIndexBlock)bm.pinBlock(header.getRoot(), loadFactory);
    }

    /**
     * Create a new root for the B+Tree, having [[left]] and [[right]] as its
     * children.
     */
    private void handleRootSplit(BTIndexBlock.SplitInformation split, BTIndexBlock origin) throws IOException, BufferManagerException, DiskManagerException {
        BTIndexBlockInner newRoot = bm.newBlock(innerFactory, BTIndexBlockInner.class);
        newRoot.initializeAsRoot(split.getSplitKey(), origin, split.getNewBlock());
        setRoot(newRoot);
        bm.unpinBlock(newRoot);
    }

    /**
     * Ensures that the schema of the given key matches our internal schema
     */
    private void checkSchema(Record key) throws SchemaException{
        if (!this.keySchema.equals(key.getSchema()))
            throw new SchemaException("Schema of the record differs from that specified in the block");
    }

    private void setSize(int size)  { header.setSize(size); }
    public  int  getSize()          { return header.getSize(); }

    /**
     * Add the specified record to the index, splitting the root of the B+Tree
     * if necessary.
     */
    public void addRecord(Record key, RecordAddress address)
        throws DuplicateKeyException, SchemaException, IOException, DiskManagerException, BufferManagerException {
        checkSchema(key);

        BTIndexBlock root = pinRoot();
        BTIndexBlock.SplitInformation split = root.addRecord(key, address, this.bm);

        if (split.splitted()) {
            handleRootSplit(split, root);
            bm.unpinBlock(split.getNewBlock());
        }

        bm.unpinBlock(root);

        setSize(getSize() + 1);
    }

    /**
     * Add the specified record to the index, splitting the root of the B+Tree
     * if necessary.
     */
    public BTIndexIterator findRecord(Record key)
        throws SchemaException, IOException, DiskManagerException, BufferManagerException {
        checkSchema(key);

        BTIndexBlock root = pinRoot();
        BTIndexIterator result = root.findRecord(bm, key);
        bm.unpinBlock(root);

        return result;
    }

    /**
     * Destroy the B+Tree
     */
    public void destroy() {
        try {
            BTIndexBlock root = pinRoot();
            root.destroy(bm);
            bm.freeBlock(root);
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    public String toString() {
        String result = "";
        try {
            BTIndexBlock root = pinRoot();
            result = root.toString(bm);
            bm.unpinBlock(root);
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
        return result;
    }
}
