/*-------------------------------------------------------------------
   Authors: 2003: Stijn Vansummeren

   The Schema class provides capabilities to describe a relation
   schema.

   $Id: Schema.java,v 1.1.1.1 2005/09/26 08:19:41 stijn Exp $
-------------------------------------------------------------------*/
package be.ulb.db.schema;

import java.util.*;

public class Schema
{
    protected Dictionary<String,Integer> name_to_index; // A mapping from attributes name to indices
    protected String [] index_to_name;                  // A mapping from attribute indices to names
    protected AttributeType [] types;                   // The type of each attribute
    protected int bytesize = 0;                         // number of bytes needed to store a record adhering to this schema
    protected int [] offsets;                           // The offset of each attribute of this schema when stored on disk

    /** Construct a schema consisting of a sequence of (name, type) tuples.
     *
     * @param names The name of each attribute in this schema (in order).
     * @param type The type of each attribute in this schema (in order).
     */
    public Schema(String [] names, AttributeType [] types) throws SchemaException
    {
        initSchema(names, types);
    }

    /** Initialize the schema contained in this object. The various attributes name
     * and types must be specified, and sanity-checks are performed.
     *
     * @param names The name of each attribute in this schema (in order).
     * @param type The type of each attribute in this schema (in order).
     */
    protected void initSchema(String [] names, AttributeType [] types) throws SchemaException
    {
        if (names.length != types.length )
            throw new SchemaException ("Names and types parameters should be of equal length!");

        this.types = new AttributeType[types.length];
        this.name_to_index = new Hashtable<String,Integer>();
        this.index_to_name = new String[types.length];
        this.offsets = new int[types.length];

        for (int i = 0; i < types.length; i++)
            {
                this.types[i] = types[i];
                offsets[i] = bytesize;
                bytesize += types[i].bytesize();

                /* we do not allow a null-name */
                if (names[i] == null)
                    throw new SchemaException("Attribute names cannot be null!");

                /* store it in the name_to_index mapping to allow name resolving */
                if (this.name_to_index.put(names[i], new Integer(i)) != null)
                    throw new SchemaException("Attribute names should be unique!");

                /* also store the name in index_to_name to allow
                 * reverse-lookup */
                this.index_to_name[i] = names[i];
            }
    }

    /** Default constructor is only available to subclasses.
     * Subclasses must ensure that they call [[initSchema]] explicitly in their
     * constructor. */
    protected Schema() { }

    /** Returns the index of the attribute named [[name] */
    public int resolveName(String name) throws SchemaException
    {
        Integer attrindex = name_to_index.get(name);
        if (attrindex == null)
            throw new SchemaException("Attribute with name " + name + " not found");

        return attrindex.intValue();
    }

    /** Returns the name of the attribute at [[index]] */
    public String getName(int index) throws SchemaException
    {
        try {
            return index_to_name[index];
        }
        catch(ArrayIndexOutOfBoundsException e)  {
            throw new SchemaException("No attribute with index " + index);
        }
    }

    /** Returns the type of attribute with of the attribute named [[name]]. */
    public AttributeType getType(String name) throws SchemaException
    {
        return getType(name_to_index.get(name));
    }

    /** Returns the type of attribute with of the attribute at [[index]]. */
    public AttributeType getType(Integer index) throws SchemaException
    {
        if (index != null)
            return getType(index.intValue());
        else
            throw new SchemaException("No attribute with index null");
    }

    /** Returns the type of attribute of the attribute at [[index]]. */
    public AttributeType getType(int index) throws SchemaException
    {
        try {
            return types[index];
        }
        catch(ArrayIndexOutOfBoundsException e)  {
            throw new SchemaException("No attribute with index " + index);
        }
    }

    /** Returns the number of bytes needed to store a record adhering
     * to this schema. */
    public int bytesize()
    {
        return bytesize;
    }

    /** Returns the offset (in bytes) the attribute at index starts at
     * in a record. */
    public int getOffset(int index)
    {
        return offsets[index];
    }

    /** Gets the number of attributes in this schema */
    public int length()
    {
        return types.length;
    }

    /** Checks whether this schema is a subschema of [[other]]. A schema is a
     * subschema of another if every attribute of the first schema occur in the
     * latter schema */
    public boolean isSubSchemaOf(Schema other)
    {
        if (this == other)
            return true;

        if (other.length() < this.length())
            return false;

        boolean subschema = true;
        for (int i = 0; i < types.length && subschema; i++) {
            Integer otherindex = other.name_to_index.get(index_to_name[i]);

            if (otherindex == null)
                subschema &= false;
            else
                subschema &= types[i].equals(other.types[otherindex.intValue()]);
        }

        return subschema;
    }

    /** Project the schema on the specified attributes.
     *
     * @attribute names The name of the attributes to keep in the resulting schema.
     *
     * @return A schema that has the attribute of this schema in the order specified
     * by [[names]].
     */
    public Schema project(String [] names) throws SchemaException
    {
        AttributeType [] types = new AttributeType[names.length];
        for (int i = 0; i < names.length; i++) {
           try {
               types[i] = this.types[(name_to_index.get(names[i])).intValue()];
           }
           catch (NullPointerException e) {
               throw new SchemaException("Scheme does not contain attribute " + names[i]);
           }
        }

        return new Schema(names,types);
    }

    /** Checks if two schema objects are semantically equal:
     * two schemas are semantically equal if they have the same number
     * of attribute, and their attributes are of the same type and of the same name and
     * are in the same order.
     *
     * @param o The other schema, to be compared for equality with this one.
     *
     * @return True if this schema is semantically equivalent to [[o]] or false
     * otherwise.
     */
    public boolean equals(Object o)
    {
        if (this == o)
            return true;

        Schema other = (Schema) o;

        if (other.length() != this.length())
            return false;

        boolean equal = true;
        for (int i = 0; i < length() && equal; i++) {
            equal &= types[i].equals(other.types[i]);
            equal &= index_to_name[i].equals(other.index_to_name[i]);
        }

        return equal;
    }

    public int hashCode()
    {
        int code = 0;

        for (int i = 0; i < index_to_name.length; i++)
            code += index_to_name[i].hashCode();

        return code / index_to_name.length;
    }

    /**
     * Creates a schema with unnamed attributes from the given [[types]].
     * Attributes names are attributed randomly, while ensuring that there is no
     * clash.
     *
     * @param types The type of the attributes that constitute the schema (in
     * order)
     *
     * @return A schema where each attribute is typed according to [[types]]
     */
    public static Schema createAnonymousSchema(AttributeType... types) {
        String[] names = new String[types.length];
        Schema result = null;
        for (int i = 0; i != names.length; ++i) {
            names[i] = "_" + i;
        }

        try {
            result = new Schema(names, types); }
        catch (SchemaException e) {
            /* Occurs if names.length != types.length
             * This should never be the case.
             */
            e.printStackTrace();
            System.exit(1);
        }

        return result;
    }
}

